@@ -116,6 +116,7 @@ static const struct mach_header*	sMainExecutableMachHeader = NULL;
 static cpu_type_t					sHostCPU;
 static cpu_subtype_t				sHostCPUsubtype;
 static ImageLoader*					sMainExecutable = NULL;
+static bool							sAllImagesMightContainUnlinkedImages;	 // necessary until will support dylib unloading
 static std::vector<ImageLoader*>	sAllImages;
 static std::vector<ImageLoader*>	sImageRoots;
 static std::vector<ImageLoader*>	sImageFilesNeedingTermination;
@@ -652,7 +653,7 @@ void processDyldEnvironmentVarible(const char* key, const char* value)
 	}
 }
 
-static void checkEnvironmentVariables(const char* envp[])
+static void checkEnvironmentVariables(const char* envp[], bool ignoreEnviron)
 {
 	const char* home = NULL;
 	const char** p;
@@ -660,7 +661,7 @@ static void checkEnvironmentVariables(const char* envp[])
 		const char* keyEqualsValue = *p;
 	    if ( strncmp(keyEqualsValue, "DYLD_", 5) == 0 ) {
 			const char* equals = strchr(keyEqualsValue, '=');
-			if ( equals != NULL ) {
+			if ( (equals != NULL) && !ignoreEnviron ) {
 				const char* value = &equals[1];
 				const int keyLen = equals-keyEqualsValue;
 				char key[keyLen+1];
@@ -741,15 +742,36 @@ bool validImage(ImageLoader* possibleImage)
 
 uint32_t getImageCount()
 {
-	return sAllImages.size();
+	if ( sAllImagesMightContainUnlinkedImages ) {
+		uint32_t count = 0;
+		for (std::vector<ImageLoader*>::iterator it=sAllImages.begin(); it != sAllImages.end(); it++) {
+			if ( (*it)->isLinked() )
+				++count;
+		}
+		return count;
+	}
+	else {
+		return sAllImages.size();
+	}
 }
 
 ImageLoader* getIndexedImage(unsigned int index)
 {
-	if ( index < sAllImages.size() )
-		return sAllImages[index];
-	else
-		return NULL;
+	if ( sAllImagesMightContainUnlinkedImages ) {
+		uint32_t count = 0;
+		for (std::vector<ImageLoader*>::iterator it=sAllImages.begin(); it != sAllImages.end(); it++) {
+			if ( (*it)->isLinked() ) {
+				if ( index == count )
+					return *it;
+				++count;
+			}
+		}
+	}
+	else {
+		if ( index < sAllImages.size() )
+			return sAllImages[index];
+	}
+	return NULL;
 }
 
 ImageLoader* findImageByMachHeader(const struct mach_header* target)
@@ -1152,7 +1174,7 @@ static ImageLoader* loadPhase6(int fd, struct stat& stat_buf, const char* path,
 			pread(fd, firstPage, 4096, fileOffset);
 		}
 		else {
-			throw "no matching architecture in fat wrapper";
+			throw "no matching architecture in universal wrapper";
 		}
 	}
 	
@@ -1435,8 +1457,8 @@ static ImageLoader* loadPhase1(const char* path, const LoadContext& context, std
 	if ( image != NULL )
 		return image;
 	
-	// try fallback paths
-	if ( (sEnv.DYLD_FALLBACK_FRAMEWORK_PATH != NULL) || (sEnv.DYLD_FALLBACK_LIBRARY_PATH != NULL) ) {
+	// try fallback paths during second time (will open file)
+	if ( (exceptions != NULL) && ((sEnv.DYLD_FALLBACK_FRAMEWORK_PATH != NULL) || (sEnv.DYLD_FALLBACK_LIBRARY_PATH != NULL)) ) {
 		image = loadPhase2(path, context, sEnv.DYLD_FALLBACK_FRAMEWORK_PATH, sEnv.DYLD_FALLBACK_LIBRARY_PATH, exceptions);
 		if ( image != NULL )
 			return image;
@@ -1561,6 +1583,20 @@ ImageLoader* cloneImage(ImageLoader* image)
 
 ImageLoader* loadFromMemory(const uint8_t* mem, uint64_t len, const char* moduleName)
 {
+	// if fat wrapper, find usable sub-file
+	const fat_header* memStartAsFat = (fat_header*)mem;
+	uint64_t fileOffset = 0;
+	uint64_t fileLength = len;
+	if ( memStartAsFat->magic == OSSwapBigToHostInt32(FAT_MAGIC) ) {
+		if ( fatFindBest(memStartAsFat, &fileOffset, &fileLength) ) {
+			mem = &mem[fileOffset];
+			len = fileLength;
+		}
+		else {
+			throw "no matching architecture in universal wrapper";
+		}
+	}
+
 	// try mach-o each loader
 	if ( isCompatibleMachO(mem) ) {
 		ImageLoader* image = new ImageLoaderMachO(moduleName, (mach_header*)mem, len, gLinkContext);
@@ -1829,7 +1865,23 @@ static void setContext(int argc, const char* argv[], const char* envp[], const c
 	gLinkContext.apple					= apple;
 }
 
-
+static bool checkEmulation()
+{
+#if __i386__
+	int mib[] = { CTL_KERN, KERN_CLASSIC, getpid() };
+	int is_classic = 0;
+	size_t len = sizeof(int);
+	int ret = sysctl(mib, 3, &is_classic, &len, NULL, 0);
+	if ((ret != -1) && is_classic) {
+		// When a 32-bit ppc program is run under emulation on an Intel processor,
+		// we want any i386 dylibs (e.g. the emulator) to not load in the shared region
+		// because the shared region is being used by ppc dylibs
+		gLinkContext.sharedRegionMode = ImageLoader::kDontUseSharedRegion;
+		return true;
+	}
+#endif
+	return false;
+}
 
 void link(ImageLoader* image, ImageLoader::BindingLaziness bindness, ImageLoader::InitializerRunning runInitializers)
 {
@@ -1851,7 +1903,13 @@ void link(ImageLoader* image, ImageLoader::BindingLaziness bindness, ImageLoader
 	}
 
 	// process images
-	image->link(gLinkContext, bindness, runInitializers, sAddImageCallbacks.size());
+	try {
+		image->link(gLinkContext, bindness, runInitializers, sAddImageCallbacks.size());
+	}
+	catch (const char* msg) {
+		sAllImagesMightContainUnlinkedImages = true;
+		throw msg;
+	}
 	
 #if OLD_GDB_DYLD_INTERFACE
 	// notify gdb that loaded libraries have changed
@@ -1885,7 +1943,8 @@ _main(const struct mach_header* mainExecutableMH, int argc, const char* argv[],
 	}
 	uintptr_t result = 0;
 	sMainExecutableMachHeader = mainExecutableMH;
-	checkEnvironmentVariables(envp);
+	bool isEmulated = checkEmulation();
+	checkEnvironmentVariables(envp, isEmulated);
 	if ( sEnv.DYLD_PRINT_OPTS ) 
 		printOptions(argv);
 	if ( sEnv.DYLD_PRINT_ENV ) 
